
/* 
 * Copyright 2010 Carsten Clauss, Chair for Operating Systems,
 *                                RWTH Aachen University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <string.h>
#include <stdio.h>

#include "RCCE.h"
#include "iRCCE.h"

#undef  _CACHE_WARM_UP_
#undef  _ERROR_CHECK_
#undef  _USE_WILDCARDS_
#undef  _USE_RCCE_API_

#define MAXBUFSIZE  1024*1024*4
#define DEFAULTLEN  1024
#define NUMROUNDS   10000

char buffer[MAXBUFSIZE+1];
char dummy = 0;

#ifdef _OPENMP
#pragma omp threadprivate (buffer, dummy)
#endif

#ifdef  _USE_WILDCARDS_
#ifndef iRCCE_ANY_LENGTH
#ifdef SINGLEBITFLAGS
#warning iRCCE_ANY_LENGTH: for using this wildcard, SINGLEBITFLAGS must be disabled! (make SINGLEBITFLAGS=0)
#endif
#ifdef RCCE_VERSION
#warning iRCCE_ANY_LENGTH: for using this wildcard, iRCCE must be built againt RCCE release V1.0.13!
#endif
#define iRCCE_ANY_LENGTH length
#endif
#endif

int RCCE_APP(int argc, char **argv)
{
  int i;
  int num_ranks;
  int root_rank = 0;
  int my_rank;
  int numrounds = NUMROUNDS;
  int maxlen    = DEFAULTLEN;
  int length;
  int round;
  double timer;

  RCCE_init(&argc, &argv);
  iRCCE_init();

  my_rank   = RCCE_ue();
  num_ranks = RCCE_num_ues();

  if(argc > 1) numrounds = atoi(argv[1]);

  if(numrounds < 1)
  {
    if(my_rank == 0) fprintf(stderr, "Multicast needs at least 1 round; try again\n");
    exit(-1);
  }

  if(argc > 2) maxlen = atoi(argv[2]);

  if(maxlen < 1)
  {
    if(my_rank == 0) fprintf(stderr, "Illegal message size: %s; try again\n", argv[2]);
    exit(-1);
  }
  else if(maxlen > MAXBUFSIZE)
  {
    if(my_rank == 0) fprintf(stderr, "Message size %d is too big; try again\n", maxlen);
    exit(-1);  
  }
  
  if(argc > 3) root_rank = atoi(argv[3]);

  if(root_rank < 0 || root_rank >= num_ranks)
  {
    if(my_rank == 0) fprintf(stderr, "Illegal root rank: %s; try again\n", argv[3]);
    exit(-1);
  }
  
  if(my_rank == 0) printf("\n[MULTI-CAST] root is %d / size is %d\n", root_rank, num_ranks);
#ifdef _CACHE_WARM_UP_
  if(my_rank == 0) printf("* CACHE_WARM_UP\n");
#endif
#ifdef _ERROR_CHECK_
  if(my_rank == 0) printf("* ERROR_CHECK\n");
#endif
#ifdef _USE_WILDCARDS_
  if(my_rank == 0) printf("* USE_WILDCARDS\n");
#else
#ifdef _USE_RCCE_API_
  if(my_rank == 0) printf("* USE_RCCE_API\n");
#endif
#endif

  if(my_rank == 0) printf("#bytes\t\tusec\t\tMB/sec\n");

  for(length=1; length <= maxlen; length*=2)
  {

#ifdef _CACHE_WARM_UP_
    for(i=0; i < length; i++)
    {
      /* cache warm-up: */
      dummy += send_buffer[i];
      dummy += recv_buffer[i];
    }
#endif

    /* synchronize before starting PING-PONG: */
    RCCE_barrier(&RCCE_COMM_WORLD);
  
    for(round=0; round < numrounds+1; round++)
    {
      if(my_rank == root_rank)
      {
#ifdef _ERROR_CHECK_     
	for(i=0; i < length; i++)
	{
	  buffer[i] = (i+length+round) % 127;
	}
#endif

#ifdef _USE_RCCE_API_
	RCCE_bcast(buffer, length, my_rank, RCCE_COMM_WORLD);
#else
	iRCCE_msend(buffer, length);
#endif
      } 
      else
      {
#ifdef _USE_RCCE_API_
	RCCE_bcast(buffer, length, root_rank, RCCE_COMM_WORLD);
#else
#ifdef  _USE_WILDCARDS_
	iRCCE_mrecv(buffer, iRCCE_ANY_LENGTH, iRCCE_ANY_SOURCE);
#else
	iRCCE_mrecv(buffer, length, root_rank);
#endif
#endif
      }

      /* start timer: */
      if(round == 0) timer = RCCE_wtime();

#ifdef _ERROR_CHECK_
      for(i=0; i < length; i++)
      {
	if(buffer[i] != (i+length+round) % 127)
	{
	  fprintf(stderr, "ERROR: %d VS %d\n", buffer[i], (i+length+round) % 127);
	  exit(-1);
	}
      }
#endif
    }
    
    /* stop timer: */
    timer = RCCE_wtime() - timer;
    
    if(my_rank == 0) {
      printf("%d\t\t%1.2lf\t\t%1.2lf\n", length, timer/(2.0*numrounds)*1000000, (length / (timer/(2.0*numrounds))) / (1024*1024) );
      fflush(stdout);
    }
  }

  RCCE_finalize();

  return 0;
}

