
/* 
 * Copyright 2010 Carsten Clauss, Chair for Operating Systems,
 *                                RWTH Aachen University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <string.h>
#include <stdio.h>

#include "RCCE.h"
#include "RCCE_lib.h"
#include "iRCCE.h"

#undef _ERROR_CHECK_
#undef _USE_SEPARATED_BUFFERS_

#ifdef SINGLEBITFLAGS
#define iRCCE_MAX_TAGGED_LEN 0
#endif
#define TAGBUFSIZE  iRCCE_MAX_TAGGED_LEN
#define NUMROUNDS   10000

#ifdef _USE_SEPARATED_BUFFERS_
char send_buffer[TAGBUFSIZE];
char recv_buffer[TAGBUFSIZE];
#else
#define send_buffer buffer
#define recv_buffer buffer
char buffer[TAGBUFSIZE];
#endif

int RCCE_APP(int argc, char **argv)
{
  int i;
  int num_ranks;
  int remote_rank, my_rank;
  int numrounds = NUMROUNDS;
  int length = TAGBUFSIZE;
  int round;
  double timer;
  RCCE_FLAG mailbox;

  RCCE_init(&argc, &argv);
  iRCCE_init();

  my_rank   = RCCE_ue();
  num_ranks = RCCE_num_ues();

  if(num_ranks != 2)
  {
    if(my_rank == 0) fprintf(stderr, "This program needs exactly two UEs; try again\n");
    exit(-1);
  }


  if(argc > 1) numrounds = atoi(argv[1]);

  if(numrounds < 1)
  {
    numrounds = NUMROUNDS;    
  }

  if(argc > 2) length = atoi(argv[2]);

  if(length < 0)
  {
    length = TAGBUFSIZE;    
  }

#ifdef SINGLEBITFLAGS
#warning THIS CODE CAN _NOT_ BE USED WITH SINGLEBITFLAGS!
  if(my_rank == 0) printf("\nTHIS CODE CAN _NOT_ BE USED WITH SINGLEBITFLAGS!\n");
#else

  if(my_rank == 0) printf("\n[TAGGED-FLAGS]\n");

#ifdef _ERROR_CHECK_
  if(my_rank == 0) printf("* ERROR_CHECK\n");
#endif
#ifdef _USE_SEPARATED_BUFFERS_
  if(my_rank == 0) printf("* USE_SEPARATED_BUFFERS\n");
#endif

  remote_rank = (my_rank + 1) % 2;

  for(i=0; i<TAGBUFSIZE; i++)
  {
    send_buffer[i] = i;
    recv_buffer[i] = 0;
  }

  iRCCE_flag_alloc_tagged(&mailbox);

  iRCCE_flag_write_tagged(&mailbox, RCCE_FLAG_UNSET, remote_rank, NULL, 0);

  // synchronize before starting:
  RCCE_barrier(&RCCE_COMM_WORLD);
  
  for(round=0; round < numrounds+1; round++)
  {
    if(my_rank == 0)
    {
      // Ping:
      iRCCE_flag_write_tagged(&mailbox, RCCE_FLAG_SET, remote_rank, send_buffer, length);
      // Pong:
      iRCCE_wait_tagged(mailbox, RCCE_FLAG_SET, recv_buffer, length);
      iRCCE_flag_write(&mailbox, RCCE_FLAG_UNSET, my_rank);

      // start timer:
      if(round==0) timer = RCCE_wtime();

#ifdef _ERROR_CHECK_
      for(i=0; i<length; i++)
      {
	if(recv_buffer[i] != i)
	{
	  fprintf(stderr, "ERROR: %d VS %d\n", recv_buffer[i], i);
	  exit(-1);
	}
	recv_buffer[i] = 0;
      }
#endif
    }
    else
    {
      // Ping:
      iRCCE_wait_tagged(mailbox, RCCE_FLAG_SET, recv_buffer, length);
      iRCCE_flag_write(&mailbox, RCCE_FLAG_UNSET, my_rank);
      // Pong:
      iRCCE_flag_write_tagged(&mailbox, RCCE_FLAG_SET, remote_rank, send_buffer, length);
    }
  }

  // stop timer:
  timer = RCCE_wtime() - timer;
    
  if(my_rank == 0) {
    printf("length = %d byte / latency = %1.2lf us\n", length, timer/(2.0*numrounds)*1000000);
    fflush(stdout);
  }    

#endif
  
  RCCE_finalize();

  return 0;
}

