
/* 
 * Copyright 2010 Carsten Clauss, Chair for Operating Systems,
 *                                RWTH Aachen University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 *  This is quite similar to the common Ping-Pong/Ping-Ping benchmarks.
 *  However, here we are using the iRCCE_ANY_LENGTH/iRCCE_ANY_SOURCE
 *  wildcards when receiving the messages.
 */

#include <string.h>
#include <stdio.h>

#include "RCCE.h"
#include "iRCCE.h"

#undef  _ERROR_CHECK_

#define MAXBUFSIZE  1024*1024*4
#define DEFAULTLEN  1024
#define NUMROUNDS   10000

char send_buffer[MAXBUFSIZE+1];
char recv_buffer[MAXBUFSIZE+1];
char dummy = 0;

#ifdef _OPENMP
#pragma omp threadprivate (send_buffer, send_buffer, dummy)
#endif

#ifndef iRCCE_ANY_LENGTH
#ifdef SINGLEBITFLAGS
#warning iRCCE_ANY_LENGTH: for using this wildcard, SINGLEBITFLAGS must be disabled! (make SINGLEBITFLAGS=0)
#endif
#ifdef RCCE_VERSION
#warning iRCCE_ANY_LENGTH: for using this wildcard, iRCCE must be built againt RCCE release V1.0.13!
#endif
#define iRCCE_ANY_LENGTH length
#endif

int RCCE_APP(int argc, char **argv)
{
  int i;
  int num_ranks;
  int remote_rank, my_rank;
  int numrounds = NUMROUNDS;
  int maxlen    = DEFAULTLEN;
  int length;
  int round;
  double timer;
  iRCCE_SEND_REQUEST send_request;
  iRCCE_RECV_REQUEST recv_request;

  RCCE_init(&argc, &argv);
  iRCCE_init();

  my_rank   = RCCE_ue();
  num_ranks = RCCE_num_ues();

  if(argc > 1) numrounds = atoi(argv[1]);

  if(numrounds < 1)
  {
    if(my_rank == 0) fprintf(stderr, "Wildcards needs at least 1 round; try again\n");
    exit(-1);
  }

  if(argc > 2) maxlen = atoi(argv[2]);

  if(maxlen < 1)
  {
    if(my_rank == 0) fprintf(stderr, "Illegal message size: %s; try again\n", argv[2]);
    exit(-1);
  }
  else if(maxlen > MAXBUFSIZE)
  {
    if(my_rank == 0) fprintf(stderr, "Message size %d is too big; try again\n", maxlen);
    exit(-1);
  }

  if(num_ranks != 2)
  {
    if(my_rank == 0) fprintf(stderr, "Wildcards needs exactly two UEs; try again\n");
    exit(-1);
  }

  remote_rank = (my_rank + 1) % 2;

  if(my_rank == 0) printf("\n[PING-PONG] ...using wildcards!\n");
  if(my_rank == 0) printf("#bytes\t\tusec\t\tMB/sec\n");

  for(length=1; length <= maxlen; length*=2)
  {
    /* synchronize before starting PING-PONG: */
    RCCE_barrier(&RCCE_COMM_WORLD);

    for(round=0; round < numrounds+1; round++)
    {

#ifdef _ERROR_CHECK_
      for(i=0; i < length; i++)
      {
	send_buffer[i] = (i+length+round) % 127;
      }
#endif

      if(my_rank == 0)
      {
	/* send PING via pipelined blocking send: */
	iRCCE_ssend(send_buffer, length, remote_rank);
	/* recv PONG via pipelined blocking recv: */
	iRCCE_srecv(recv_buffer, iRCCE_ANY_LENGTH, remote_rank);
      } 
      else
      {
	/* recv PING via pipelined blocking recv: */
	iRCCE_srecv(recv_buffer, iRCCE_ANY_LENGTH, remote_rank);
	/* send PONG via pipelined blocking send: */
	iRCCE_ssend(send_buffer, length, remote_rank);
      }

      /* start timer: */
      if(round == 0) timer = RCCE_wtime();

#ifdef _ERROR_CHECK_
      for(i=0; i < length; i++)
      {
	if(recv_buffer[i] != (i+length+round) % 127)
	{
	  fprintf(stderr, "ERROR: %d VS %d\n", recv_buffer[i], (i+length+round) % 127);
	  exit(-1);
	}
      }
#endif
    }
    
    /* stop timer: */
    timer = RCCE_wtime() - timer;
    
    if(my_rank == 0) printf("%d\t\t%1.2lf\t\t%1.2lf\n", length, timer/(2.0*numrounds)*1000000, (length / (timer/(2.0*numrounds))) / (1024*1024) );
  }
 
  if(my_rank == 0) printf("\n[PING-PING] ...using wildcards!\n");
  if(my_rank == 0) printf("#bytes\t\tusec\t\tMB/sec\n");

  for(length=1; length <= maxlen; length*=2)
  {    
    /* synchronize before starting PING-PING: */
    RCCE_barrier(&RCCE_COMM_WORLD);
    
    for(round=0; round < numrounds+1; round++)
    {

#ifdef _ERROR_CHECK_
      {
	int i;
	for(i=0; i < length; i++)
	{
	  send_buffer[i] = (i+length+round) % 127;
	}
      }
#endif

      /* send PING via non-blocking send: */
      iRCCE_isend(send_buffer, length, remote_rank, &send_request);

      /* recv PING via non-blocking recv: */
      iRCCE_irecv(recv_buffer, iRCCE_ANY_LENGTH, remote_rank, &recv_request);

     
      /* wait for completion: */
      iRCCE_isend_wait(&send_request);
      iRCCE_irecv_wait(&recv_request);

      if( iRCCE_get_length() != length ) 
      {
	fprintf(stderr, "ERROR: length %d VS %d\n", iRCCE_get_length(), length );
	exit(-1);	
      }
      
      /* start timer: */
      if(round==0) timer = RCCE_wtime();

#ifdef _ERROR_CHECK_
      {
	int i;
	for(i=0; i < length; i++)
	{
	  if(recv_buffer[i] != (i+length+round) % 127)
	  {
	    fprintf(stderr, "ERROR: %d VS %d\n", recv_buffer[i], (i+length+round) % 127);
	    exit(-1);
	  }
	}
      }
#endif
    }
    
    /* stop timer: */
    timer = RCCE_wtime() - timer;
    
    if(my_rank == 0) printf("%d\t\t%1.2lf\t\t%1.2lf\n", length, timer/(numrounds)*1000000, (length / (timer/(numrounds))) / (1024*1024) );
  }

  RCCE_finalize();

  return 0;
}

